"""
Custom bo4e object, that was created for a migration project
"""

from decimal import Decimal
from typing import Optional

from bo4e.com.com import COM
from bo4e.enum.mengeneinheit import Mengeneinheit


class Zaehlpunkt(COM):
    """
    The zaehlpunkt object was created during a migration project.
    It contains attributes needed for metering mapping.
    """

    periodenverbrauch_vorhersage: Decimal
    einheit_vorhersage: Mengeneinheit = Mengeneinheit.KWH
    zeitreihentyp: str = "Z21"
    kunden_wert: Decimal | None
    einheit_kunde: Optional[Mengeneinheit] = None
    grundzustaendiger: bool = True



"""
Contains the dataset for the resource loader.
It also contains the validation logic for the resource loader dataset.
"""
from bo4e.bo.marktlokation import Marktlokation
from bo4e.bo.messlokation import Messlokation
from bo4e.bo.vertrag import Vertrag
from bo4e.bo.zaehler import Zaehler

from ibims.datasets import DataSetBaseModel


class TripicaResourceLoaderDataSet(DataSetBaseModel):
    """
    This is a bo4e data set that consists of
    * a Vertrag
    * a Marktlokation
    * a Messlokation
    * a Zaehler.
    In the context of this package is may be used to create Tripica Resources.
    """

    marktlokation: Marktlokation
    messlokation: Messlokation
    vertrag: Vertrag
    zaehler: Zaehler



"""
Bundles extended BOs
"""
from ibims.bo.bilanzierung import Bilanzierung
from ibims.bo.dokument import Dokument
from ibims.bo.file import File
from ibims.bo.geschaeftspartner import GeschaeftspartnerErweitert
from ibims.bo.hinweis import Hinweis
from ibims.bo.kampagne import Kampagne
from ibims.bo.marktlokation_erweitert import MarktlokationErweitert
from ibims.bo.rechnung import RechnungErweitert
from ibims.bo.zaehler import ZaehlerErweitert, ZaehlerGas



"""
extension of the official BO4E marktlokation
"""
from typing import Literal

from bo4e.bo.marktlokation import Marktlokation

from ibims.enum import MesstechnischeEinordnung, Regelzone


class MarktlokationErweitert(Marktlokation):
    """
    extension of the official BO4E marktlokation
    """

    messtechnische_einordnung: MesstechnischeEinordnung
    uebertragungsnetzgebiet: Regelzone
    variant: Literal["Z14", "Z15"]
    """
        gridUsageBilling.variant (Netznutzungsabrechnungsvariante)
            Z14: WorkPriceBasicPrice (Arbeitspreis/Grundpreis)
            Z15: WorkPricePerformancePrice (Arbeitspreis/Leistungspreis)
    """
    community_id: str



"""
New ENUM to differentiate note types.
"""
from bo4e.enum.strenum import StrEnum


class Regelzone(StrEnum):
    """
    a strenum where the str-value of each member is the ILN of the Regelzone
    """

    TRANSNETBW = "10YDE-ENBW-----N"
    TENNETDE = "10YDE-EON------1"
    AMPRION = "10YDE-RWENET---I"
    FUENFZIGHERTZ = "10YDE-VE-------2"
    GAS = "37Y005053MH0000R"



"""
contains the BO model for kampagne
"""
from bo4e.bo.geschaeftsobjekt import Geschaeftsobjekt


class Kampagne(Geschaeftsobjekt):
    """
    A "Kampagne"/campaign models which marketing activities led customers to a product/tariff.
    """

    id: str  #: eindeutige (eher technische) bezeichnung der kampagne
    name: str  #: nicht unbedingt eindeutige aber menschenlesbare Bezeichnung e.g. "Social Media Kampagne Q3 2022



"""
Create a generic reference to a file object
"""
from typing import Optional

from pydantic import BaseModel


class File(BaseModel):
    """
    This class represents a file that is stored in the database.
    """

    file_name_for_docstore: Optional[str] = None
    """
    The name of the file in the docstore
    """

    folder_name_for_docstore: Optional[str] = None
    """
    The folder name of the file in the docstore
    """

    file: bytes
    """
    The body of the file
    """



"""
Extend BO Geschäftspartner to hold various more information.
"""
from datetime import datetime
from typing import Optional

from bo4e.bo.geschaeftspartner import Geschaeftspartner


class GeschaeftspartnerErweitert(Geschaeftspartner):
    """
    enhance the businesspartner object
    """

    # see https://github.com/bo4e/BO4E-python/issues/443

    # optional attributes
    erstellungsdatum: Optional[datetime] = None
    """
    when was the businesspartner created
    """
    geburtstag: Optional[datetime] = None
    """
    When was the Business Partner born
    """
    telefonnummer_mobil: Optional[str] = None
    """
    Telephone-number for communications
    """
    telefonnummer_privat: Optional[str] = None
    """
    Telephone number for communications (landline)
    """
    telefonnummer_geschaeft: Optional[str] = None
    """
    Telephone number if the customer has a business
    """
    firmenname: Optional[str] = None
    """
    The name of the company if the GeschaeftspartnerErweitert represents an company
    """
    # added firmenname because name1, name2, name3 are already for surname, firstname and title
    # don't use gewerbekennzeichnung for mapping logic of company name,
    # because in database the company name is not filled all the time.

    hausbesitzer: Optional[bool] = None
    """
    Flag if the bussinespartner is the House owner.
    It is None if the Flag is not set by any reason
    """



"""
Extend BO Zaehler to hold information about the data acquisition method.
New BO ZaehlerGas to disambiguate between use case `Strom` and `Gas`.
"""
from datetime import datetime
from typing import Literal, Optional

from bo4e.bo.zaehler import Zaehler
from bo4e.com.zeitraum import Zeitraum
from bo4e.enum.geraetemerkmal import Geraetemerkmal
from bo4e.enum.netzebene import Netzebene
from pydantic import field_validator

from ibims.enum import Messwerterfassung, ZaehlerTypErweitert


class ZaehlerErweitert(Zaehler):
    """
    Extend `Zaehler` with `Messwerterfassung`
    """

    zaehlertyp: ZaehlerTypErweitert  # type: ignore[assignment]

    messwerterfassung: Messwerterfassung

    nachstes_ablesedatum: Optional[datetime] = None

    aktiver_zeitraum: Optional[Zeitraum] = None


class ZaehlerGas(ZaehlerErweitert):
    """
    Resolve some ambiguity of `Strom` and `Gas`
    """

    zaehlergroesse: Geraetemerkmal
    """
    Add information 'Zählergröße' for 'Gaszähler'.
    """

    druckniveau: Literal[Netzebene.ND, Netzebene.MD, Netzebene.HD]

    @field_validator("zaehlergroesse")
    @classmethod
    def validate_zaehlergroesse(cls, zaehlergroesse: Geraetemerkmal) -> Geraetemerkmal:
        """
        Thanks to the ambiguity in the bo4e model we need to check if the `Geraetemerkmal` is indeed a 'Zählergröße'.
        """
        if not zaehlergroesse.value.startswith("GAS_"):
            raise ValueError(f"Illegal value for 'zaehlergroesse': {zaehlergroesse}")
        return zaehlergroesse



"""
the Hochfrequenz Bilanzierungsobjekt
"""
from datetime import datetime
from typing import Optional

from bo4e.bo.geschaeftsobjekt import Geschaeftsobjekt

from ibims.enum import BoTypErweitert


class Bilanzierung(Geschaeftsobjekt):
    """
    Bilanzierung is a business object used for balancing. This object is no BO4E standard and a complete go
    implementation can be found at
    https://github.com/Hochfrequenz/go-bo4e/blob/3414a1eac741542628df796d6beb43eaa27b0b3e/bo/bilanzierung.go
    """

    bo_typ: BoTypErweitert = BoTypErweitert.BILANZIERUNG  # type: ignore[assignment]

    bilanzierungsbeginn: datetime
    """
    inclusive start date of balancing
    """
    bilanzierungsende: datetime
    """
    exclusive end date of balancing
    """

    bilanzkreis: Optional[str] = None



"""
extension of the official BO4E Artikelnummer
"""
# pylint:disable=line-too-long
from enum import StrEnum


class BDEWArtikelnummerErweitert(StrEnum):
    """
    Extension of BDEW Artikelnummern. The additional fields can be found in the go implementation:
    https://github.com/Hochfrequenz/go-bo4e/blob/8dda93f8eda51557bb355a93e94c379111f0242b/enum/bdewartikelnummer/bdewartikelnummer.go
    """

    LEISTUNG = "LEISTUNG"  #: Leistung
    LEISTUNG_PAUSCHAL = "LEISTUNG_PAUSCHAL"  #: Leistung pauschal
    GRUNDPREIS = "GRUNDPREIS"  #: Grundpreis
    REGELENERGIE_ARBEIT = "REGELENERGIE_ARBEIT"  #: Regelenergie Arbeit
    REGELENERGIE_LEISTUNG = "REGELENERGIE_LEISTUNG"  #: Regelenergie Leistung
    NOTSTROMLIEFERUNG_ARBEIT = "NOTSTROMLIEFERUNG_ARBEIT"  #: Notstromlieferung Arbeit
    NOTSTROMLIEFERUNG_LEISTUNG = "NOTSTROMLIEFERUNG_LEISTUNG"  #: Notstromlieferung Leistung
    RESERVENETZKAPAZITAET = "RESERVENETZKAPAZITAET"  #: Reservenetzkapazität
    RESERVELEISTUNG = "RESERVELEISTUNG"  #: Reserveleistung
    ZUSAETZLICHE_ABLESUNG = "ZUSAETZLICHE_ABLESUNG"  #: Zusätzliche Ablesung
    PRUEFGEBUEHREN_AUSSERPLANMAESSIG = "PRUEFGEBUEHREN_AUSSERPLANMAESSIG"  #: Prüfgebühren (außerplanmäßig)
    WIRKARBEIT = "WIRKARBEIT"  #: Wirkarbeit
    SINGULAER_GENUTZTE_BETRIEBSMITTEL = (
        "SINGULAER_GENUTZTE_BETRIEBSMITTEL"  #: singulär genutzte Betriebsmittel (z. B. Trafomiete, Leitungen)
    )
    ABGABE_KWKG = "ABGABE_KWKG"  #: Abgabe KWKG
    ABSCHLAG = "ABSCHLAG"  #: Abschlag
    KONZESSIONSABGABE = "KONZESSIONSABGABE"  #: Konzessionsabgabe
    ENTGELT_FERNAUSLESUNG = "ENTGELT_FERNAUSLESUNG"  #: Entgelt für Fernauslesung
    UNTERMESSUNG = "UNTERMESSUNG"  #: Untermessung
    BLINDMEHRARBEIT = "BLINDMEHRARBEIT"  #: Blindmehrarbeit
    ENTGELT_ABRECHNUNG = "ENTGELT_ABRECHNUNG"  #: Entgelt für Abrechnung
    SPERRKOSTEN = "SPERRKOSTEN"  #: Sperrkosten
    ENTSPERRKOSTEN = "ENTSPERRKOSTEN"  #: Entsperrkosten
    MAHNKOSTEN = "MAHNKOSTEN"  #: Mahnkosten
    MEHR_MINDERMENGEN = "MEHR_MINDERMENGEN"  #: Mehr- und Mindermenge
    INKASSOKOSTEN = "INKASSOKOSTEN"  #: Inkassokosten
    BLINDMEHRLEISTUNG = "BLINDMEHRLEISTUNG"  #: Blindmehrleistung
    ENTGELT_MESSUNG_ABLESUNG = "ENTGELT_MESSUNG_ABLESUNG"  #: Entgelt für Messung und Ablesung
    ENTGELT_EINBAU_BETRIEB_WARTUNG_MESSTECHNIK = (
        "ENTGELT_EINBAU_BETRIEB_WARTUNG_MESSTECHNIK"  #: Entgelt für Einbau, Betrieb und Wartung der Messtechnik
    )
    AUSGLEICHSENERGIE = "AUSGLEICHSENERGIE"  #: Ausgleichsenergie
    ZAEHLEINRICHTUNG = "ZAEHLEINRICHTUNG"  #: Zähleinrichtung
    WANDLER_MENGENUMWERTER = "WANDLER_MENGENUMWERTER"  #: Wandler/Mengenumwerter
    KOMMUNIKATIONSEINRICHTUNG = "KOMMUNIKATIONSEINRICHTUNG"  #: Kommunikationseinrichtung
    TECHNISCHE_STEUEREINRICHTUNG = "TECHNISCHE_STEUEREINRICHTUNG"  #: Technische Steuereinrichtung
    PARAGRAF_19_STROM_NEV_UMLAGE = "PARAGRAF_19_STROM_NEV_UMLAGE"  #: § 19 StromNEV Umlage
    BEFESTIGUNGSEINRICHTUNG = "BEFESTIGUNGSEINRICHTUNG"  #: Befestigungseinrichtung (z. B. Zählertafel)
    OFFSHORE_HAFTUNGSUMLAGE = "OFFSHORE_HAFTUNGSUMLAGE"  #: Offshore-Haftungsumlage
    FIXE_ARBEITSENTGELTKOMPONENTE = "FIXE_ARBEITSENTGELTKOMPONENTE"  #: Fixe Arbeitsentgeltkomponente
    FIXE_LEISTUNGSENTGELTKOMPONENTE = "FIXE_LEISTUNGSENTGELTKOMPONENTE"  #: Fixe Leistungsentgeltkomponente
    UMLAGE_ABSCHALTBARE_LASTEN = "UMLAGE_ABSCHALTBARE_LASTEN"  #: Umlage abschaltbare Lasten
    MEHRMENGE = "MEHRMENGE"  #: Mehrmenge
    MINDERMENGE = "MINDERMENGE"  #: Mindermenge
    ENERGIESTEUER = "ENERGIESTEUER"  #: Energiesteuer
    SMARTMETER_GATEWAY = "SMARTMETER_GATEWAY"  #: Smartmeter-Gateway
    STEUERBOX = "STEUERBOX"  #: Steuerbox
    MSB_INKL_MESSUNG = "MSB_INKL_MESSUNG"  #: Messtellenbetrieb inklusive Messung
    AUSGLEICHSENERGIE_UNTERDECKUNG = "AUSGLEICHSENERGIE_UNTERDECKUNG"  #: AusgleichsenergieUnterdeckung
    ZUSATZDIENSTLEISTUNG_PARAGRAPH_35_2_1_MSBG = "ZUSATZDIENSTLEISTUNG_PARAGRAPH_35_2_1_MSBG"  #: Zusatzdienstleistung nach § 35 Abs. 2 Nr. 1 MsbG with article number 9990001000813
    ZUSATZDIENSTLEISTUNG_PARAGRAPH_35_2_2_MSBG = "ZUSATZDIENSTLEISTUNG_PARAGRAPH_35_2_2_MSBG"  #: Zusatzdienstleistung nach § 35 Abs. 2 Nr. 2 MsbG with article number 9990001000821
    ZUSATZDIENSTLEISTUNG_PARAGRAPH_35_2_3_MSBG = "ZUSATZDIENSTLEISTUNG_PARAGRAPH_35_2_3_MSBG"  #: Zusatzdienstleistung nach § 35 Abs. 2 Nr. 3 MsbG with article number 9990001000839
    ZUSATZDIENSTLEISTUNG_PARAGRAPH_35_2_4_MSBG = "ZUSATZDIENSTLEISTUNG_PARAGRAPH_35_2_4_MSBG"  #: Zusatzdienstleistung nach § 35 Abs. 2 Nr. 4 MsbG with article number 9990001000847
    ZUSATZDIENSTLEISTUNG_PARAGRAPH_35_2_5_MSBG = "ZUSATZDIENSTLEISTUNG_PARAGRAPH_35_2_5_MSBG"  #: Zusatzdienstleistung nach § 35 Abs. 2 Nr. 5 MsbG with article number 9990001000855
    ZUSATZDIENSTLEISTUNG_PARAGRAPH_35_3_MSBG = "ZUSATZDIENSTLEISTUNG_PARAGRAPH_35_3_MSBG"  #: Zusatzdienstleistung nach § 35 Abs. 3 MsbG with article number 9990001000863



"""
This a docstring for the module.
"""


class MyClass:  # pylint: disable=too-few-public-methods
    """
    This is a docstring for the class.
    """

    def __init__(self):
        """
        Initialize for the sake of initializing
        """
        self.my_instance_var: str = "abc"

    def do_something(self) -> str:
        """
        Actually does nothing.
        :return: the value of an instance variable
        """
        # this is a super long line with: 100 < line length <= 120 to demonstrate the purpose of pyproject.toml
        return self.my_instance_var



"""
This module contains hints for contracts or customers
"""
from datetime import datetime

from bo4e.bo.geschaeftsobjekt import Geschaeftsobjekt

from ibims.enum import HinweisThema


class Hinweis(Geschaeftsobjekt):
    """
    Contains specific hints for the handling of contracts and customers.
    Hints are meant to be read and written by agents or customer service employees.
    """

    erstellungsdatum: datetime  #: when the note has been created

    thema: HinweisThema | str

    nachricht: str  #: the note itself; e.g. 'Hat Hotline angerufen; Beschwert sich über zu hohen Abschlag'



"""
Messtechnische Einordnung ENUM
"""

from bo4e.enum.strenum import StrEnum


class MesstechnischeEinordnung(StrEnum):
    """
    An enum for the messtechnische Einordnung
    """

    IMS = "IMS"  #: Intelligentes Messsysteme
    KME_MME = "KME_MME"  #: Konventionelle Messeinrichtung / Moderne Messeinrichtung
    KEINE_MESSUNG = "KEINE_MESSUNG"



"""
Bundles new enum
"""
from ibims.enum.abgabeart import Abgabeart
from ibims.enum.bdewartikelnummer import BDEWArtikelnummerErweitert
from ibims.enum.botyp import BoTypErweitert
from ibims.enum.hinweisthema import HinweisThema
from ibims.enum.messtechnische_einordnung import MesstechnischeEinordnung
from ibims.enum.messwerterfassung import Messwerterfassung
from ibims.enum.messwertstatus import Messwertstatus
from ibims.enum.rechnung_erweitert import RechnungstypErweitert
from ibims.enum.regelzone import Regelzone
from ibims.enum.zaehlertyp_erweitert import ZaehlerTypErweitert



"""
Extend BO rechnung to hold more possible invoice types
"""
from typing import Optional

from bo4e.bo.rechnung import Rechnung

from ibims.com import RechnungspositionErweitert
from ibims.enum import RechnungstypErweitert


class RechnungErweitert(Rechnung):
    """
    replace enum Rechnungstyp mit RechnungstypErweitert and add attributes, which can be found in the
    go implementation
    https://github.com/Hochfrequenz/go-bo4e/blob/3414a1eac741542628df796d6beb43eaa27b0b3e/bo/rechnung.go
    """

    rechnungspositionen: Optional[list[RechnungspositionErweitert]] = None  # type: ignore[assignment]
    rechnungstyp: RechnungstypErweitert  # type: ignore[assignment]
    ist_selbstausgestellt: Optional[bool] = None
    ist_reverse_charge: Optional[bool] = None



"""
A Preisgarantie with MetaData added
"""
from typing import Optional

from bo4e.com.preisgarantie import Preisgarantie
from bo4e.enum.preisgarantietyp import Preisgarantietyp

from ibims.meta import MetaDataMixin


class PreisgarantieErweitert(Preisgarantie, MetaDataMixin):
    """
    eine Preisgarantie mit Metadaten
    """

    preisgarantietyp: Optional[Preisgarantietyp] = None  # type:ignore[assignment]
    """overrides the original preisgarantietyp with a nullable"""



"""
Contains the dataset for the invoice loader
"""
from typing import Optional

from bo4e.bo.marktlokation import Marktlokation
from bo4e.bo.marktteilnehmer import Marktteilnehmer
from bo4e.bo.vertrag import Vertrag

from ibims.bo import Bilanzierung, RechnungErweitert
from ibims.datasets import DataSetBaseModel
from ibims.targetmodels import TransaktionsdatenInvoices


class InvoiceLoaderDataSet(DataSetBaseModel):
    """
    This is a bo4e dat set that consists out of
    * a Marktlokation
    * a RechnungErweitert
    * a Marktteilnhemer
    * a Bilanzierung (which is only relevant for a memi)
    * a Vertrag (which is only relevant for a memi)
    """

    marktlokation: Marktlokation
    """
    The following attributes need to be filled for this DataSet:
    - lokationsadresse
    - marktlokations_id
    - sparte
    """

    rechnung_erweitert: RechnungErweitert
    """
    The following attributes need to be filled for this Dataset:
    - rechungstyp
    - rechungsnummer
    - rechungsdatum
    - faelligkeitsdatum
    - rechungsperiode
    - gesamtnetto
    - gesamtsteuer
    - gesamtbrutto, this value should be negative for self issued invoices
    - zuzahlen, this value should be negative for self issued invoices
    - rechnungspositionen, the values of the fields teilsumme_steuer and teilsumme_netto should be negative 
        for self issued invoices
    - steuerbetrag, the values of the fields basiswert and steuerwert should be negative for self issued invoices
    - ist_selbstausgestellt,
    - ist_reverse_charge
    """

    marktteilnehmer: Marktteilnehmer
    """
    The following attributes need to be filled for this DataSet:
    - name1 (organization)
    - partneradresse
    - rollencodenummer
    - rollencodetyp
    - ansprechoartner
    - sparte
    """

    bilanzierung: Optional[Bilanzierung] = None
    """
    The following attributes need to be filled for this DataSet:
    - bilanzierungsbeginn
    - bilanzierungsende
    """

    vertrag: Optional[Vertrag] = None
    """
    The following attributes need to be filled for this DataSet:
    - vertragskonditionen (only the field netznutzungsabbrechung is needed)
    """

    transaktionsdaten: TransaktionsdatenInvoices
    """
    The following attributes need to be filled for this DataSet:
    - sparte
    - pruefidentifikator
    - lieferrichtung
    - duplikat
    - referenznummer
    - datenaustauschreferenz
    - absender
    - empfaenger
    - nachrichtendatum
    - nachrichten_referenznummer
    """



"""
New COM for holding information about sepa mandates
"""
from datetime import datetime
from typing import Optional

from bo4e.com.com import COM


class SepaInfo(COM):
    """
    This class includes details about the sepa mandates.
    """

    # required attributes

    sepa_id: str
    """
    System internal id of the SEPA mandate.
    """
    # muss noch überprüft werden wie es allgemeingültig umzusetzen ist.

    sepa_zahler: bool
    """
    there may be sepa information regardless of whether it is used or not. this field
    confirms the use case e.g. it is false if the customer pays for himself
    """

    # optional attributes
    creditor_identifier: Optional[str] = None
    """
    Creditor Identifier is a number that identify the creditor of the sepa transaction
    """

    gueltig_seit: Optional[datetime] = None
    """
    Inklusiver Zeitpunkt ab dem das SEPA Mandat gültig ist.
    """



"""
Contains the dataset for the quantities loader
"""
from bo4e.bo.energiemenge import Energiemenge
from bo4e.bo.marktlokation import Marktlokation
from bo4e.bo.marktteilnehmer import Marktteilnehmer
from bo4e.bo.messlokation import Messlokation

from ibims.datasets.base import DataSetBaseModel
from ibims.targetmodels import TransaktionsdatenQuantities


class QuantitiesLoaderDataSet(DataSetBaseModel):
    """
    This is a bo4e dat set that consists out of
    * a Energiemenge
    * a Messlokation or Marktlokation
    * a Marktteilnehmer
    """

    energiemenge: Energiemenge
    """
    The following attributes need to be filled for this DataSet:
    - lokations_id
    - lokationstyp
    - energieverbrauch
    """

    lokation: Messlokation | Marktlokation
    """
    The following attributes need to be filled for this DataSet:
    - messlokations_id/marktlokations_id
    - sparte
    """

    sender: Marktteilnehmer
    """
    The following attributes need to be filled for this DataSet:
    - gewerbekennzeichnung
    - rollencodenummer
    - rollencodetyp
    """

    empfaenger: Marktteilnehmer
    """
    The following attributes need to be filled for this DataSet:
    - gewerbekennzeichnung
    - rollencodenummer
    - rollencodetyp
    """

    transaktionsdaten: TransaktionsdatenQuantities



"""
Extension of the official steuerbetrag
"""
from decimal import Decimal
from typing import Optional

from bo4e.com.steuerbetrag import Steuerbetrag


class SteuerbetragErweitert(Steuerbetrag):
    """
    This class extends the com Steuerbetrag with the attribute steuerwert_vorausgezahlt, which can also be found in the
    go implementation:
    https://github.com/Hochfrequenz/go-bo4e/blob/8dda93f8eda51557bb355a93e94c379111f0242b/com/steuerbetrag.go
    """

    steuerwert_vorausgezahlt: Optional[Decimal] = None  #: tax amount of the prepaid amounts



"""
Extend COM Verbrauch to hold various more information
"""

from datetime import datetime
from decimal import Decimal
from typing import Optional

from bo4e.com.verbrauch import Verbrauch
from bo4e.enum.strenum import StrEnum

from ibims.enum import Messwertstatus


class AblesendeRolle(StrEnum):
    """
    Eine (Markt)Rolle, die Verbräuche abliest.
    """

    VNB = "VNB"  #: der verteilnetzbetreiber
    ENDKUNDE = "ENDKUNDE"  #: der Endkunde selbst
    VORIGER_LIEFERANT = "VORIGER_LIEFERANT"  #: der vorherige Lieferant hat einen Verbrauch mitgeteilt)
    MSB = "MSB"  #: der messstellenbetreiber
    SYSTEM = "SYSTEM"  #: wert ist eine vom system erstellte schätzung


class Ablesungsstatus(StrEnum):
    """
    State of the reading
    """

    GUELTIG = "GUELTIG"
    UNGUELTIG = "UNGUELTIG"
    ABGERECHNET = "ABGERECHNET"


class VerbrauchErweitert(Verbrauch):
    """
    enhance the consumption object
    """

    # see https://github.com/bo4e/BO4E-python/issues/443

    # optional attributes
    ablesegrund: Optional[str] = None
    """
    Reason why the counter is read
    """
    ablesebeschreibung: Optional[str] = None
    """
    Details to the reading
    """
    periodenverbrauch: Optional[Decimal] = None

    periodenverbrauch_ursprung: Optional[str] = None

    ableser: Optional[AblesendeRolle] = None
    """
    from whom is thew reading coming from
    """
    status: Optional[Ablesungsstatus] = None
    """
    reading status
    """
    energiegehalt_gas: Optional[Decimal] = None

    energiegehalt_gas_gueltig_von: Optional[datetime] = None
    """
    valid from date for the calorific amount (inclusive start)
    """
    energiegehalt_gas_gueltig_bis: Optional[datetime] = None
    """
    valid to date for the calorific amount (exclusive end)
    """

    umwandlungsfaktor_gas: Optional[Decimal] = None

    umwandlungsfaktor_gas_gueltig_von: Optional[datetime] = None
    """
    valid from date for the conversion factor (inclusive start)
    """
    umwandlungsfaktor_gas_gueltig_bis: Optional[datetime] = None
    """
    valid to date for the conversion factor (exclusive end)
    """
    messwertstatus: Optional[Messwertstatus] = None
    """
    the messwertstatus is the status of meter reading
    """

    def is_single_tariff(self) -> bool:
        """
        Returns true if the consumption is single tariff ("Eintarif")
        """
        return "1.8.0" in self.obis_kennzahl

    def is_high_tariff(self) -> bool:
        """
        Returns true if the consumption is high tariff (HT, "Hochtarif")
        """
        return "1.8.1" in self.obis_kennzahl

    def is_low_tariff(self) -> bool:
        """
        Returns true if the consumption is low tariff (NT, "Niedertarif")
        """
        return "1.8.2" in self.obis_kennzahl



"""
Extend COM Zählwerk to hold various more information
"""

from datetime import datetime
from typing import Optional

from bo4e.com.zaehlwerk import Zaehlwerk

from ibims.enum import Abgabeart


class ZaehlwerkErweitert(Zaehlwerk):
    """
    enhance the register object
    """

    # see https://github.com/bo4e/BO4E-python/issues/443

    # optional attributes
    vorkommastellen: int
    """
    Integer places of the register
    """
    nachkommastellen: int
    """
    Decimal places of the register
    """
    schwachlastfaehig: bool

    konzessionsabgaben_typ: Optional[Abgabeart] = None

    active_from: datetime

    active_until: Optional[datetime] = None

    description: Optional[str] = None

    verbrauchsart: Optional[str] = None



"""
Bundles extended and new COMs
"""
from ibims.com.adresse import AdresseErweitert
from ibims.com.bankverbindung import Bankverbindung
from ibims.com.concessionfee import ConcessionFee
from ibims.com.preisgarantie import PreisgarantieErweitert
from ibims.com.preisposition import PreispositionErweitert
from ibims.com.rechnungsposition import RechnungspositionErweitert
from ibims.com.sepa_info import SepaInfo
from ibims.com.steuerbetrag import SteuerbetragErweitert
from ibims.com.verbrauch import VerbrauchErweitert
from ibims.com.vertragskonto import Vertragskonto, VertragskontoCBA, VertragskontoMBA
from ibims.com.zaehlpunkt import Zaehlpunkt
from ibims.com.zaehlwerk import ZaehlwerkErweitert



"""
Contains the dataset for the usage loader.
It also contains the validation logic for the usage loader dataset.
"""
from bo4e.bo.energiemenge import Energiemenge
from bo4e.bo.messlokation import Messlokation
from bo4e.bo.zaehler import Zaehler
from bomf.model import Bo4eDataSet


class TripicaUsageLoaderDataSet(Bo4eDataSet):
    """
    This is a bo4e data set that consists of
    * an Energiemenge
    * a Messlokation
    * a Zaehler.
    In the context of this package it contains all information to construct
    the tripica loader model.
    """

    energiemenge: Energiemenge
    messlokation: Messlokation
    zaehler: Zaehler



"""
Create a generic reference to a document
"""
from datetime import datetime

from bo4e.bo.geschaeftsobjekt import Geschaeftsobjekt


class Dokument(Geschaeftsobjekt):
    """
    A generic document reference like for bills, order confirmations and cancellations
    """

    erstellungsdatum: datetime
    """
    when was the document created
    """
    has_been_sent: bool
    """
    true iff the document was sent to the customer
    """
    dokumentenname: str
    """
    name of the document
    """
    vorlagenname: str
    """
    template of a document
    """



"""
extension for the official bo4e adresse
"""

from typing import Optional

from bo4e.com.adresse import Adresse


class AdresseErweitert(Adresse):
    """
    Extend Adresse with attribute ortsteil
    """

    ortsteil: Optional[str] = None



"""
extension of the official BO4E Rechnungsposition
"""

from typing import Optional

from bo4e.com.rechnungsposition import Rechnungsposition

from ibims.enum import BDEWArtikelnummerErweitert


class RechnungspositionErweitert(Rechnungsposition):
    """
    This class extend the com Rechnungsposition by replacing the enum BDEWArtikelnummer
    with BDEWArtikelNummerErweitert
    """

    artikelnummer: Optional[BDEWArtikelnummerErweitert] = None  # type: ignore[assignment]



"""
New ENUM for concession fee type
"""

from bo4e.enum.strenum import StrEnum


class Abgabeart(StrEnum):
    """
    This AbgabeArt models the Konzessionsabgabentyp.
    It contains concessionfee types needed for concessionFee mapping.
    """

    # See https://github.com/Hochfrequenz/BO4E-dotnet/blob/main/BO4E/ENUM/AbgabeArt.cs

    KAS = "KAS"
    SA = "SA"
    SAS = "SAS"
    TA = "TA"
    TAS = "TAS"
    TK = "TK"
    TKS = "TKS"
    TS = "TS"
    TSS = "TSS"



"""
Custom bo4e object, that was created for a migration project
"""

from datetime import datetime

from bo4e.com.com import COM


class ConcessionFee(COM):
    # https://github.com/Hochfrequenz/integrated-bo4e-migration-models/issues/7
    """
    The Concession Fee object was created during a migration project.
    It contains attributes needed for metering mapping.
    """

    market_location_id: str
    """market location id of powercloud db"""
    group: str | None = None
    """group of the concession fee (e.g. "TA")"""
    obis: str
    """obis code"""
    active_from: datetime
    """inclusive active_from"""
    active_until: datetime | None = None
    """exclusive active_until, None means it is still active"""
    ka: str | None = None
    """amount of concession fee"""



"""
New COM to hold information about banking data
"""
from datetime import datetime
from typing import Optional

from bo4e.com.com import COM
from pydantic import constr, field_validator

from ibims.com.sepa_info import SepaInfo


class Bankverbindung(COM):
    """
    This component contains bank connection information.
    """

    # required attributes

    iban: Optional[constr(min_length=15, max_length=34)] = None  # type:ignore[valid-type]
    """
    The IBAN (International Bank Account Number) is structured as follows:
    # 2-digit: country code
    # 2-digit: check digits
    # up to 30-digit: account identification

    """
    # international variiert die Länge, länge(DE) = 22

    bic: Optional[str] = None
    """
    The BIC (Business Identifier Code) consists 8 or 11 alphanumeric characters, structured as follows:
    # 4-digit: bank code
    # 2-digit: country code
    # 2-digit: location code
    # 3-digit: branch office (optional)
    """

    # optional attributes
    gueltig_seit: Optional[datetime] = None
    """
    Inclusive date from which on the account information is valid
    """
    gueltig_bis: Optional[datetime] = None
    """
    Inclusive date till which on the account information is valid
    """
    bankname: Optional[str] = None  #: Name der Bank, z.b. 'Sparkasse Bremen'
    """
    Consists the name of the Bank.
    """
    sepa_info: Optional[SepaInfo] = None  #: Informationen über das SEPA-Mandant
    """
    contains an object of the SepaInfo class that contains details about the sepa mandates.
    """
    kontoinhaber: Optional[str] = None
    """
    contains the name of the account holder in the format 'firstname lastname'
    """
    ouid: int
    """
    contains the ouid that is need for the paymentMeans in the Customer Loader
    """

    # pylint:disable=unused-argument, no-self-argument
    @field_validator("iban")
    @classmethod
    def validate_iban_laenge(cls, iban: str) -> str:
        """
        validate the length of the iban.
        """
        return iban
        # validierung auf ländercod und sonst nur zahlen noch bauen.
        # vlaidierung der iban nach https://ibanvalidieren.de/verifikation.html noch einbauen
        # bzw. https://de.wikipedia.org/wiki/Internationale_Bankkontonummer#Zusammensetzung international.

    @field_validator("bic")
    @classmethod
    def validate_bic_laenge(cls, bic: str) -> str:
        """
        validate the length of the bic.
        """
        if len(bic) == 8 or len(bic) == 11:
            return bic
        raise ValueError("bic has not 8 or 11 characters.")

    # validierung auf alphanumerisch noch bauen



"""
Contains the dataset for the network loader.
It also contains the validation logic for the network loader dataset.
"""
from typing import Optional

from bo4e.bo.marktteilnehmer import Marktteilnehmer
from bo4e.bo.vertrag import Vertrag
from bo4e.com.adresse import Adresse

from ibims.bo import Bilanzierung, GeschaeftspartnerErweitert, MarktlokationErweitert, ZaehlerErweitert
from ibims.com import ConcessionFee, Zaehlpunkt
from ibims.datasets import DataSetBaseModel


class TripicaNetworkLoaderDataSet(DataSetBaseModel):
    """
    This is a bo4e data set that consists of
    * two GeschaeftspartnerErweitert as customer
    * a Adresse
    * a Marktlokation
    * a Messlokation
    * two Marktteilnehmer
    * a Zaehler
    In the context of this package is may be used to create Tripica Network Data.
    """

    kunde: GeschaeftspartnerErweitert
    """
    The following attributes need to be filled for this DataSet and is used for
    the customer that holds the contract:
    - name1 (Surname)
    - name2 (Firstname)
    - name3 (Title e.g. Dr.)
    - gewerbekennzeichnung
    - partneradresse
    """

    liefer_adresse: Adresse

    geschaeftspartner_mit_rechnungs_adresse: GeschaeftspartnerErweitert
    """
    This information is used for the customerMeterReadings part and
    is not a real business partner, so not all information of this object arte needed:
    - name1 (Surname)
    - name2 (Firstname)
    - partneradresse
    """
    # TODO: get a new table join with firstname and surname seperated.
    # https://github.com/Hochfrequenz/powercloud2lynqtech/issues/818

    marktlokation: MarktlokationErweitert
    """
    The following attributes need to be filled for this DataSet:
    - marktlokations_id
    - sparte
    - bilanzierungsgebiet
    - unterbrechbar
    - gasqualitaet (if sparte is GAS)
    - netzbetreibercodenr
    - netzgebietsnr
    - kundengruppen
    - messtechnische_einordnung
    - ubertragungsnetzgebiet
    """

    netzbetreiber: Optional[Marktteilnehmer] = None
    """
    The following attributes need to be filled for this DataSet:
    - name1
    - rollencodetyp
    - rollencodenummer
    """

    zaehlpunkt: Zaehlpunkt

    messstellenbetreiber: Optional[Marktteilnehmer] = None
    """
    The following attributes need to be filled for this DataSet:
    - rollencodetyp
    - rollencodenummer
    """

    zaehler: ZaehlerErweitert
    """
    The following attributes need to be filled for this DataSet:
    - zaehlernummer
    - sparte
    - zaehlerauspraegung
    - zaehlertyp
    - zaehlergroeßeGas
    - tarifart
    - zaehlerauspraegung
    - zaehlwerke with each
        - zaehlwerk_id?
        - obis_kennzahl
    """

    vertrag: Vertrag
    """
    The following attributes need to be filled for this DataSet:
    - vertragsnummer    (not mapped just for traking)
    - vertragsbeginn
    - vertragsende
    """

    bilanzierung: Bilanzierung
    """
    The following attributes need to be filled for this DataSet:
    - bilanzkreis
    - bilanzierungsbeginn
    """

    concession_fees: list[ConcessionFee] | None

    slp_profilname: str | None

    tlp_profilname: str | None



"""
New COM to hold information about different variations and reasons for account locks
"""

from bo4e.com.preisposition import Preisposition
from bo4e.enum.steuerkennzeichen import Steuerkennzeichen


class PreispositionErweitert(Preisposition):
    """
    Extesion of the bo4e Preisposition with taxrate
    """

    steuersatz: Steuerkennzeichen
    """
    The taxrate is given as a enum
    """



"""
New ENUM to differentiate the data acquisition method in BO Zähler.
"""
from bo4e.enum.strenum import StrEnum


class Messwerterfassung(StrEnum):
    """
    Specify data acquisition method
    """

    FERNAUSLESBAR = "FERNAUSLESBAR"
    MANUELL_AUSGELESENE = "MANUELL_AUSGELESENE"



"""
Extension of the official bo4e rechnungstyp
"""

from bo4e.enum.strenum import StrEnum


class RechnungstypErweitert(StrEnum):
    """
    Abbildung verschiedener Rechnungstypen zur Kennzeichnung von Rechnungen.
    Die neu hinzugefügten Rechnungstypen können in der go Implementierung gefunden werden.
    """

    # pylint:disable=line-too-long
    # https://github.com/Hochfrequenz/go-bo4e/blob/3414a1eac741542628df796d6beb43eaa27b0b3e/enum/rechnungstyp/rechnungstyp.go

    ENDKUNDENRECHNUNG = "ENDKUNDENRECHNUNG"
    """Eine Rechnung vom Lieferanten an einen Endkunden über die Lieferung von Energie"""
    NETZNUTZUNGSRECHNUNG = "NETZNUTZUNGSRECHNUNG"
    """Eine Rechnung vom Netzbetreiber an den Netznutzer. (i.d.R. der Lieferant) über die Netznutzung."""
    MEHRMINDERMENGENRECHNUNG = "MEHRMINDERMENGENRECHNUNG"
    """ Eine Rechnung vom Netzbetreiber an den Netznutzer. (i.d.R. der Lieferant)
    zur Abrechnung von Mengen-Differenzen zwischen Bilanzierung und Messung."""
    MESSSTELLENBETRIEBSRECHNUNG = "MESSSTELLENBETRIEBSRECHNUNG"
    """Rechnung eines Messstellenbetreibers an den Messkunden."""
    BESCHAFFUNGSRECHNUNG = "BESCHAFFUNGSRECHNUNG"
    """Rechnungen zwischen einem  Händler und Einkäufer von Energie."""
    AUSGLEICHSENERGIERECHNUNG = "AUSGLEICHSENERGIERECHNUNG"
    """Rechnung an den Verursacher von Ausgleichsenergie."""
    TURNUSRECHNUNG = "TURNUSRECHNUNG"
    ABSCHLAGSRECHNUNG = "ABSCHLAGSRECHNUNG"
    ABSCHLUSSRECHNUNG = "ABSCHLUSSRECHNUNG"
    ZWISCHENRECHNUNG = "ZWISCHENRECHNUNG"
    INTEGRIERTE_13TE_RECHNUNG = "INTEGRIERTE_13TE_RECHNUNG"
    MONATSRECHNUNG = "MONATSRECHNUNG"
    ZUSAETZLICHE_13TE_RECHNUNG = "ZUSAETZLICHE_13TE_RECHNUNG"



"""
New COMs to model the MBA and CBA structures in the customer loader.
"""
from datetime import datetime

from bo4e.bo.vertrag import Vertrag
from bo4e.com.adresse import Adresse
from bo4e.com.com import COM
from bo4e.enum.kontaktart import Kontaktart
from pydantic import field_validator
from pydantic_core.core_schema import FieldValidationInfo


class Vertragskonto(COM):
    """
    Bundle common attributes of MBA and CBA.
    """

    # required attributes

    ouid: int
    """
    This ID is unique in a defined scope.
    In this case, it is easier to construct the ouid before the CustomerLoader mapping because our solution to
    construct these ouids involve iteration over all MBAs and CBAs. This is easier in the mapping Powercloud -> BO4E.
    """
    vertrags_adresse: Adresse
    """
    Address related to the `Vertragskonto` which is only used to create a common mba for `Vertragskonto`s with the same
    address.
    """
    vertragskontonummer: str
    """
    This `vertragskontonummer` is intended to be used for the field `name` in the billing account data
    (Customer Loader).
    """
    rechnungsstellung: Kontaktart
    """
    preferred way of delivering the bill to the customer
    """


class VertragskontoCBA(Vertragskonto):
    """
    Models a CBA (child billing account) which directly relates to a single contract. It contains information about
    locks and billing dates. But in the first place, CBAs will be grouped together by the address in their contracts.
    For each group of CBAs with a common address there will be created an MBA (master billing
    account) to support that the invoices for the CBAs can be bundled into a single invoice for the MBA.
    """

    vertrag: Vertrag
    """
    Contracts in the CBA account
    """

    erstellungsdatum: datetime
    """
    Creation date of the CBA account
    """

    rechnungsdatum_start: datetime
    """
    first billing date
    """
    rechnungsdatum_naechstes: datetime
    """
    next billing date
    """


class VertragskontoMBA(Vertragskonto):
    """
    Models an MBA (master billing account). Its main purpose is to bundle CBAs together having the same address in
    their related contracts. This feature supports a single invoice for all CBAs instead of several
    invoices for each.
    """

    cbas: list[VertragskontoCBA]

    @field_validator("cbas")
    @classmethod
    def validate_cbas_addresses(
        cls, value: list[VertragskontoCBA], info: FieldValidationInfo
    ) -> list[VertragskontoCBA]:
        """
        This validator ensures that the MBA and the related CBAs all have the same address. It also ensures that
        the MBA has at least one related CBA.
        """
        if len(value) == 0:
            raise ValueError("The MBA must have at least one CBA")
        for cba in value:
            if cba.vertrags_adresse != info.data["vertrags_adresse"]:
                raise ValueError(
                    f"The address of the cba with `vertragsnummer `{cba.vertrag.vertragsnummer} "
                    "mismatches the MBA's address."
                )
        return value



"""
Extension of the official BO4E Zaehlertyp
"""
from bo4e.enum.strenum import StrEnum


class ZaehlerTypErweitert(StrEnum):
    """Extension of the official BO4E Zaehlertyp"""

    DREHSTROMZAEHLER = "DREHSTROMZAEHLER"  #: Drehstromzaehler,

    BALGENGASZAEHLER = "BALGENGASZAEHLER"  #: Balgengaszähler,

    DREHKOLBENZAEHLER = "DREHKOLBENZAEHLER"  #: Drehkolbengaszähler,

    SMARTMETER = "SMARTMETER"  #: Smart Meter Zähler,

    LEISTUNGSZAEHLER = "LEISTUNGSZAEHLER"  #: leistungsmessender Zähler,

    MAXIMUMZAEHLER = "MAXIMUMZAEHLER"  #: Maximumzähler,

    TURBINENRADGASZAEHLER = "TURBINENRADGASZAEHLER"  #: Turbinenradgaszähler,

    ULTRASCHALLGASZAEHLER = "ULTRASCHALLGASZAEHLER"  #: Ultraschallgaszähler,

    WECHSELSTROMZAEHLER = "WECHSELSTROMZAEHLER"  #: Wechselstromzähler,

    WIRBELGASZAEHLER = "WIRBELGASZAEHLER"  #: Wirbelgaszähler,

    MESSDATENREGISTRIERGERAET = "MESSDATENREGISTRIERGERAET"  #: Messdatenregistriergerät,

    ELEKTRONISCHERHAUSHALTSZAEHLER = "ELEKTRONISCHERHAUSHALTSZAEHLER"  #: elektronischer Haushaltszähler,

    SONDERAUSSTATTUNG = "SONDERAUSSTATTUNG"  #: Individuelle Abstimmung (Sonderausstattung),

    MODERNEMESSEINRICHTUNG = "MODERNEMESSEINRICHTUNG"  #: moderne Messeinrichtung nach MsbG

    LASTGANGZAEHLER = "LASTGANGZAEHLER"  #: Lastgangzähler

    NEUEMESSEINRICHTUNG = "NEUEMESSEINRICHTUNG"  #: Neue Messeinrichtung GAS nach MsbG
